#ifndef _AUTH_H
#define _AUTH_H

#include <openssl/ssl.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/bio.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/pem.h>
#include <openssl/pkcs12.h>
#include <stdio.h>
#include <stdlib.h>

#define FILETYPE_ASN1     1
#define FILETYPE_PEM      2
#define FILETYPE_PKCS12   3

#define ERR(fmt, args...) \
  fprintf(stderr, "[%s:%d] " fmt "\n", __FILE__, __LINE__, ##args)
#define ERR_EXIT(fmt, args...) { ERR(fmt, ##args); exit(1); }
#define ERR_RET(n, fmt, args...) { ERR(fmt, ##args); return n; }
#define SSL_ERR_STRING ERR_error_string(ERR_get_error(), NULL)


void auth_init_bio_err();
void auth_free_bio_err();

// digital signature
EVP_PKEY *auth_load_privatekey(int filetype, const char *filename,
                          const char *password);
EVP_PKEY *auth_load_publickey(int filetype, char *filename, char *password);
int auth_sign_buffer(EVP_MD_CTX *ctx, EVP_PKEY *pkey, const char *buf,
                     int buflen, char *sigbuf, int *siglen);
int auth_verify_buffer(EVP_MD_CTX *ctx, EVP_PKEY *pkey, const char *buf,
                       int buflen, char *sigbuf, int siglen);
int auth_sign_file(EVP_MD_CTX *ctx, EVP_PKEY *pkey, const char *file,
                   int buflen, char *sigbuf, int *siglen);
int auth_verify_file(EVP_MD_CTX *ctx, EVP_PKEY *pkey, const char *file,
                     int buflen, char *sigbuf, int siglen);

// certificate verification
int auth_cert_verify_get_purpose_id(const char *purpose_name);
X509_STORE *auth_cert_verify_init(const char *ca_path, const char *ca_file);
int auth_verify_cert(X509_STORE *ctx, const char *filename, int purpose,
                     int vflags);

#endif /* _AUTH_H */
