#ifndef _ATTACKGRAPH_H
#define _ATTACKGRAPH_H

#include <router/ddos/edgemark/edgemark.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct _router {
  unsigned long   ip;
  double          localTrafficRate;
  struct _router* nextRouter;
  struct _router* nextConnection;
} Router;

typedef struct _markingFieldsNode {
  MarkingFields markingField;
  int frequency;
  struct _markingFieldsNode* next;
} MarkingFieldsNode;

typedef struct _routerTree {
  unsigned long        ip;
  int                  numOfBranch;
  int                  maxNumOfBranch;
  struct _routerTree** nextLevelArray; 
  double               localTrafficRate;
} RouterTree;

typedef struct _RouterTreeListNode {
    const RouterTree *treenode;
    const RouterTree *parentNode;
    struct _RouterTreeListNode* next;
} RouterTreeListNode;

MarkingFieldsNode* newMarkingFieldsNode();
void freeMarkingFieldsNode(MarkingFieldsNode* head);
MarkingFieldsNode* dupMarkingFieldsNode(const MarkingFieldsNode* node);
MarkingFieldsNode* dupMarkingFieldsNodeList(const MarkingFieldsNode* list);
MarkingFieldsNode* findMarkingFieldsNode(const MarkingFieldsNode* head,
                                         const MarkingFields fields);

int* extractPacketFrequency(const MarkingFieldsNode* head, 
                            const Router* attackPath,
                            int* sizeOfArray);
void printMarkingFieldsNodeList(const MarkingFieldsNode* list);
Router* newRouter();
Router* dupRouter(Router* ptr);
int insertRouter(Router* head, Router* pt);
int insertConnection(Router* head, Router* from, Router* to);
int haveRouter(Router* head, Router* pt);
void freeList(Router* head);
RouterTree* newRouterTree(int ip, int maxNumOfBrench);
RouterTree* dupRouterTree(const RouterTree* t);
int insertTreeConnection(RouterTree* head, const MarkingFields* node);
void freeRouterTree(RouterTree* head);
int isMarkingFieldEqual(MarkingFields* a, MarkingFields* b);
int addMarkingFields(MarkingFieldsNode* head, MarkingFields* field);
int attackGraphCon(RouterTree* head,
                   const MarkingFieldsNode* markingFieldList);
Router* attackPathCon(const RouterTree* head);

int printGraphList(Router* head);
void printTree(const RouterTree* head, const char* prefix);
int* extractPacketFrequency(const MarkingFieldsNode* head,
                            const Router* attackPath,
                            int* sizeOfArray);
unsigned long* writeTree(const RouterTree *root);
char* writeTreeStr(const RouterTree *node);

#endif /* _ATTACKGRAPH_H */
