#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/udp.h>
#include "ipsend.h"

static char n = 0;
static __u32 seqno = 0;
static __u32 lastudp = 0;
static __u32 last = ~0;
static unsigned int ipt_ipsend_target(struct sk_buff **pskb,
                                      unsigned int hooknum,
                                      const struct net_device *in,
                                      const struct net_device *out,
                                      const void *targinfo,
                                      void *userinfo)
{
  struct iphdr *iph = (*pskb)->nh.iph;
  int iphlen = iph->ihl << 2;
  struct ipsend_opt *opt;
  __u32 *data;

  if (hooknum == NF_IP_LOCAL_IN) {
    switch (iph->protocol) {
    case IPPROTO_UDP:
      data = (__u32 *) (((__u8 *) iph) + iphlen + sizeof(struct udphdr));
      lastudp = ntohl(data[0]);
      if (last != ~0) {
        if (last != lastudp)
          DEBUGP("last = %x  lastudp = %x \n", last, lastudp);
//        DEBUGP("UDP [%x]\n", lastudp);
        last = ~0;
      }
      break;

    case IPPROTO_IPSEND:
      opt = IPSEND_OPTPTR(iph);
      data = (__u32 *) IPSEND_DATAPTR(iph);
      last = ntohl(data[0]);
      //DEBUGP("IPSEND [%x] lastudp = %x\n", ntohl(opt->seqno), last);
      break;
    }

  } else {

    switch (iph->protocol) {
    case IPPROTO_UDP:
      data = (__u32 *) (((__u8 *) iph) + iphlen + sizeof(struct udphdr));
      lastudp = ntohl(data[0]);
      
      if (hooknum == NF_IP_LOCAL_OUT) {
        n++;
        if (n>40) n = 0;
        if (n==20) {
//        DEBUGP("send IPSEND [%x] lastudp = %x\n", seqno, lastudp);
          last = lastudp;
          ipsend_send_data(iph->saddr, iph->daddr, seqno, lastudp);
          seqno++;
        }
      } else {
        if (last != ~0) {
/*
  DEBUGP("UDP [%x] (%s)\n", lastudp,
  (hooknum == NF_IP_POST_ROUTING) ? "postrouting" : "");
*/
          if (last != lastudp)
            DEBUGP("last = %x  lastudp = %x \n", last, lastudp);
          last = ~0;
        }
      }
/*
  if (((n<=20) && (20-n<2)) || ((n>20) && (n-20<2)))
  DEBUGP("UDP [%x] (%s)\n", lastudp,
  (hooknum == NF_IP_POST_ROUTING) ? "postrouting" : "");
*/
      break;

    case IPPROTO_IPSEND:
      opt = IPSEND_OPTPTR(iph);
      data = (__u32 *) IPSEND_DATAPTR(iph);
//    DEBUGP("IPSEND [%x] lastudp = %x\n", ntohl(opt->seqno), ntohl(data[0]));
      break;
    }
  }

  return NF_ACCEPT;
}

static int ipt_ipsend_checkentry(const char *tablename,
                                 const struct ipt_entry *e,
                                 void *targinfo,
                                 unsigned int targinfosize,
                                 unsigned int hook_mask)
{
	if (hook_mask & ~((1 << NF_IP_LOCAL_IN) | (1 << NF_IP_LOCAL_OUT) |
                    (1 << NF_IP_FORWARD) | (1 << NF_IP_POST_ROUTING))) {
		DEBUGP("IPSEND: bad hook\n");
		return 0;
	}

	if (targinfosize != IPT_ALIGN(0)) {
		DEBUGP("IPSEND: targinfosize %u != 0\n", targinfosize);
		return 0;
	}

	return 1;
}

static int ipsend_proc_read(char *page, char **start, off_t off, int count,
                            int *eof, void *data)
{
  int len = 0;
  
  MOD_INC_USE_COUNT;

  len += sprintf(page+len, "seqno = %u\n", seqno);
  
  MOD_DEC_USE_COUNT;

  return len;
}

static int ipsend_proc_write(struct file *file, const char *buffer,
                             unsigned long count, void *data)
{
  int len = count;
  __u32 saddr, daddr;
  
  MOD_INC_USE_COUNT;

  saddr = 0x3264a8c0;
  daddr = 0x0764a8c1;
  
  DEBUGP("IPSEND: write\n");
  DEBUGP("IPSEND: %u.%u.%u.%u -> %u.%u.%u.%u\n",
         NIPQUAD(saddr), NIPQUAD(daddr));

  ipsend_send_data(saddr, daddr, 0, 0);
  
  MOD_DEC_USE_COUNT;
  
  return len;
}

static struct ipt_target ipt_ipsend_reg
= { { NULL, NULL }, "IPSEND", ipt_ipsend_target, ipt_ipsend_checkentry,
    NULL, THIS_MODULE };

static const char IPSEND_PROCFS_PATH[] = "ipsend";
extern struct proc_dir_entry *proc_net;

static int __init init(void)
{
  struct proc_dir_entry *proc;

  DEBUGP("IPSEND: init\n");  
  
  proc = create_proc_entry(IPSEND_PROCFS_PATH, S_IRUSR | S_IRGRP | S_IROTH, proc_net);
  if (!proc) {
    return -ENOMEM;
  }
  proc->read_proc = ipsend_proc_read;
  proc->write_proc = ipsend_proc_write;
  proc->owner = THIS_MODULE;
  
	return ipt_register_target(&ipt_ipsend_reg);
}

static void __exit fini(void)
{
  DEBUGP("IPSEND: exit\n");
  ipt_unregister_target(&ipt_ipsend_reg);
  remove_proc_entry(IPSEND_PROCFS_PATH, proc_net);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
