#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <lib/util.h>
#include <lib/net/rawsock.h>
#include <router/ddos/snapshot/snapshot.h>

inline __u8 ssmsg_msgtype(struct iphdr * iph, struct snapshot_opt * opt)
{
  if ((iph->protocol == IPPROTO_SNAPSHOT) &&
      (iph->ihl == ((sizeof(struct iphdr) + SNAPSHOT_OPTLEN) >> 2)) &&
      (opt->type == IPOPT_SNAPSHOT) && (opt->length == SNAPSHOT_OPTLEN)) {
    return opt->msgtype;
  }
  return 0;
}

void ssmsg_dump_iphdr(struct iphdr *iph)
{
  struct in_addr in;

  printf("[ iphdr ]\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n"
         "%-15.15s 0x%04x\n"
         "%-15.15s 0x%02x\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n"
         "%-15.15s 0x%04x\n",
         "version", iph->version,
         "ihl", iph->ihl,
         "tos", iph->tos,
         "tot_len", ntohs(iph->tot_len),
         "id", ntohs(iph->id),
         "flags", ntohs(iph->frag_off) >> 13,
         "frag_off", ntohs(iph->frag_off) & 0x1fff,
         "ttl", iph->ttl,
         "protocol", iph->protocol, "checksum", ntohs(iph->check));

  in.s_addr = iph->saddr;
  printf("%-15.15s %s\n", "saddr", inet_ntoa(in));
  in.s_addr = iph->daddr;
  printf("%-15.15s %s\n", "daddr", inet_ntoa(in));
}

static const char *snapshot_msgtype_str(__u8 msgtype)
{
  switch (msgtype) {
  case SNAPSHOT_MSGTYPE_RESET:
    return "RESET";
  case SNAPSHOT_MSGTYPE_CLEAR:
    return "CLEAR";
  case SNAPSHOT_MSGTYPE_CONTROL:
    return "CONTROL";
  case SNAPSHOT_MSGTYPE_VMAP:
    return "VMAP";
  case SNAPSHOT_MSGTYPE_RMAP:
    return "RMAP";
  case SNAPSHOT_MSGTYPE_MARKER:
    return "MARKER";
  case SNAPSHOT_MSGTYPE_SNAPSHOT:
    return "SNAPSHOT";
  case SNAPSHOT_MSGTYPE_CONTROL_ACK:
    return "CONTROL ACK";
  case SNAPSHOT_MSGTYPE_VMAP_ACK:
    return "VMAP ACK";
  case SNAPSHOT_MSGTYPE_RMAP_ACK:
    return "RMAP ACK";

  case SNAPSHOT_MSGTYPE_SMAP:
    return "SMAP";
  case SNAPSHOT_MSGTYPE_SMAP_ACK:
    return "SMAP ACK";
  case SNAPSHOT_MSGTYPE_STAT:
    return "STAT";
  }
  return "invalid";
}

void ssmsg_dump_opt(struct snapshot_opt *opt)
{
  printf("[ snapshot_opt ]\n"
         "%-15.15s %-3d (%s)\n"
         "%-15.15s %-3d (%s)\n"
         "%-15.15s %-3d (%s)\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n",
         "type", opt->type,
         (opt->type == IPOPT_SNAPSHOT) ? "correct" : "wrong",
         "length", opt->length,
         (opt->length == SNAPSHOT_OPTLEN) ? "correct" : "wrong",
         "msgtype", opt->msgtype,
         snapshot_msgtype_str(opt->msgtype),
         "padding", opt->padding, "seqno", ntohl(opt->seqno));
}

void ssmsg_dump_vmap(unsigned char *data, int datalen)
{
  struct snapshot_msghdr_vmap *hdr = (struct snapshot_msghdr_vmap *) data;
  __u32 *entry = (__u32 *) (data + sizeof(struct snapshot_msghdr_vmap));
  __u16 entrycnt = ntohs(hdr->entrycnt);
  struct in_addr in;
  int i;

  printf("[ snapshot_msg VMAP ]\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n",
         "entrycnt", entrycnt, "snapshotcnt", ntohs(hdr->snapshotcnt));

  if ((datalen - sizeof(struct snapshot_msghdr_vmap)) !=
      (entrycnt * sizeof(__u32))) {
    printf("data length error\n");
    return;
  }

  for (i = 0; i < entrycnt; i++) {
    in.s_addr = entry[i];
    printf("[%3d]   %-16.16s\n", i, inet_ntoa(in));
  }
}

void ssmsg_dump_rmap(unsigned char *data, int datalen)
{
  struct snapshot_msghdr_rmap *hdr = (struct snapshot_msghdr_rmap *) data;
  __u32 *entry = (__u32 *) (data + sizeof(struct snapshot_msghdr_rmap));
  __u16 entrycnt = ntohs(hdr->entrycnt);
  struct in_addr in;
  int i;

  printf("[ snapshot_msg RMAP ]\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n",
         "entrycnt", entrycnt,
         "connected", hdr->connected, "padding", hdr->padding);

  if ((datalen - sizeof(struct snapshot_msghdr_rmap)) !=
      (entrycnt * sizeof(__u32))) {
    printf("data length error\n");
    return;
  }

  for (i = 0; i < entrycnt; i++) {
    in.s_addr = entry[i];
    printf("[%3d]   %-16.16s\n", i, inet_ntoa(in));
  }
}

void ssmsg_dump_smap(unsigned char *data, int datalen)
{
  struct snapshot_msghdr_smap *hdr = (struct snapshot_msghdr_smap *) data;
  __u32 *entry = (__u32 *) (data + sizeof(struct snapshot_msghdr_smap));
  __u16 entrycnt = ntohs(hdr->entrycnt);
  struct in_addr in;
  int i;

  printf("[ snapshot_msg SMAP ]\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n",
         "entrycnt", entrycnt, "padding", ntohs(hdr->padding));

  if ((datalen - sizeof(struct snapshot_msghdr_smap)) !=
      (entrycnt * sizeof(__u32))) {
    printf("data length error\n");
    return;
  }

  for (i = 0; i < entrycnt; i++) {
    in.s_addr = entry[i];
    printf("[%3d]   %-16.16s\n", i, inet_ntoa(in));
  }
}

void ssmsg_dump_snapshot(unsigned char *data, int datalen)
{
  struct snapshot_msghdr_snapshot *hdr =
      (struct snapshot_msghdr_snapshot *) data;
  __u32 *entry =
      (__u32 *) (data + sizeof(struct snapshot_msghdr_snapshot));
  __u16 entrycnt = ntohs(hdr->entrycnt);
  struct in_addr in;
  int i;

  printf("[ snapshot_msg SNAPSHOT ]\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n",
         "entrycnt", entrycnt, "padding", ntohs(hdr->padding));

  if ((datalen - sizeof(struct snapshot_msghdr_snapshot)) !=
      (entrycnt * sizeof(__u32) * 2)) {
    printf("data length error\n");
    return;
  }

  for (i = 0; i < entrycnt; i++) {
    in.s_addr = entry[2 * i];
    printf("[%3d]   %-16.16s  %u\n", i, inet_ntoa(in),
           ntohl(entry[2 * i + 1]));
  }
}

void ssmsg_dump_stat(unsigned char *data, int datalen)
{
  struct snapshot_msghdr_stat *hdr = (struct snapshot_msghdr_stat *) data;
  __u32 *entry = (__u32 *) (data + sizeof(struct snapshot_msghdr_stat));
  __u16 entrycnt = ntohs(hdr->entrycnt);
  struct in_addr in;
  int i;

  printf("[ snapshot_msg STAT ]\n"
         "%-15.15s %d\n"
         "%-15.15s %d\n",
         "entrycnt", entrycnt, "padding", ntohs(hdr->padding));

  if ((datalen - sizeof(struct snapshot_msghdr_stat)) !=
      (entrycnt * sizeof(__u32) * 2)) {
    printf("data length error\n");
    return;
  }

  for (i = 0; i < entrycnt; i++) {
    in.s_addr = entry[2 * i];
    printf("[%3d]   %-16.16s  %u\n", i, inet_ntoa(in),
           ntohl(entry[2 * i + 1]));
  }
}

void ssmsg_dump_msg(struct snapshot_opt *opt, unsigned char *data,
                    int datalen)
{
  switch (opt->msgtype) {
  case SNAPSHOT_MSGTYPE_VMAP:
    ssmsg_dump_vmap(data, datalen);
    break;
  case SNAPSHOT_MSGTYPE_RMAP:
    ssmsg_dump_rmap(data, datalen);
    break;
  case SNAPSHOT_MSGTYPE_SMAP:
    ssmsg_dump_smap(data, datalen);
    break;
  case SNAPSHOT_MSGTYPE_SNAPSHOT:
    ssmsg_dump_snapshot(data, datalen);
    break;
  case SNAPSHOT_MSGTYPE_STAT:
    ssmsg_dump_stat(data, datalen);
    break;
  }
}

int ssmsg_send_msg(__u32 localaddr, __u32 remoteaddr, __u8 msgtype,
                   __u32 seqno, unsigned char *data, int data_len)
{
  rawsock_t raw;
  struct snapshot_opt *opt;

  memset(&raw, 0, sizeof(rawsock_t));

  if (rawsock_create_socket(&raw, IPPROTO_SNAPSHOT) < 0)
    ERR_RET(-1);

  // IP options
  opt = (struct snapshot_opt *) &(raw.opt);
  opt->type = IPOPT_SNAPSHOT;
  opt->length = SNAPSHOT_OPTLEN;
  opt->msgtype = msgtype;
  opt->padding = 0;
  opt->seqno = htonl(seqno);

  raw.data = data;
  raw.data_len = data_len;

  // fill up source and destination IP addresses
  raw.src_addr.s_addr = localaddr;
  raw.dst_addr.s_addr = remoteaddr;

  // complete packet info
  rawsock_fill_info(&raw);

  rawsock_send(&raw);

  close(raw.sd);

  return 0;
}

int ssmsg_send_vmap(__u32 localaddr, __u32 remoteaddr,
                    __u16 snapshotcnt, __u32 * ssdata, int ssdata_cnt)
{
  const int msghdr_len = sizeof(struct snapshot_msghdr_vmap);
  struct snapshot_msghdr_vmap *hdr;
  unsigned char *data;
  int data_len;
  int ret;

  data_len = msghdr_len + ssdata_cnt * sizeof(__u32);
  data = malloc(data_len);
  if (data == NULL)
    return -1;

  // VMAP message header
  hdr = (struct snapshot_msghdr_vmap *) (data);
  hdr->entrycnt = htons(ssdata_cnt);
  hdr->snapshotcnt = htons(snapshotcnt);

  // VMAP message content
  if (ssdata_cnt != 0) {
    memcpy(data + msghdr_len, ssdata, data_len - msghdr_len);
  }

  ret = ssmsg_send_msg(localaddr, remoteaddr, SNAPSHOT_MSGTYPE_VMAP,
                       0, data, data_len);

  free(data);

  return ret;
}

int ssmsg_send_rmap(__u32 localaddr, __u32 remoteaddr,
                    __u8 connected, __u32 * ssdata, int ssdata_cnt)
{
  const int msghdr_len = sizeof(struct snapshot_msghdr_rmap);
  struct snapshot_msghdr_rmap *hdr;
  unsigned char *data;
  int data_len;
  int ret;

  data_len = msghdr_len + ssdata_cnt * sizeof(__u32);
  data = malloc(data_len);
  if (data == NULL)
    return -1;

  // RMAP message header
  hdr = (struct snapshot_msghdr_rmap *) (data);
  hdr->entrycnt = htons(ssdata_cnt);
  hdr->connected = connected;
  hdr->padding = 0;

  // RMAP message content
  if (ssdata_cnt != 0) {
    memcpy(data + msghdr_len, ssdata, data_len - msghdr_len);
  }

  ret = ssmsg_send_msg(localaddr, remoteaddr, SNAPSHOT_MSGTYPE_RMAP,
                       0, data, data_len);

  free(data);

  return ret;
}

int ssmsg_send_smap(__u32 localaddr, __u32 remoteaddr,
                    __u32 * ssdata, int ssdata_cnt)
{
  const int msghdr_len = sizeof(struct snapshot_msghdr_smap);
  struct snapshot_msghdr_smap *hdr;
  unsigned char *data;
  int data_len;
  int ret;

  data_len = msghdr_len + ssdata_cnt * sizeof(__u32);
  data = malloc(data_len);
  if (data == NULL)
    return -1;

  // SMAP message header
  hdr = (struct snapshot_msghdr_smap *) (data);
  hdr->entrycnt = htons(ssdata_cnt);
  hdr->padding = 0;

  // SMAP message content
  if (ssdata_cnt != 0) {
    memcpy(data + msghdr_len, ssdata, data_len - msghdr_len);
  }

  ret = ssmsg_send_msg(localaddr, remoteaddr, SNAPSHOT_MSGTYPE_SMAP,
                       0, data, data_len);

  free(data);

  return ret;
}

int ssmsg_send_reset(__u32 localaddr, __u32 remoteaddr)
{
  return ssmsg_send_msg(localaddr, remoteaddr, SNAPSHOT_MSGTYPE_RESET,
                        0, NULL, 0);
}

int ssmsg_send_clear(__u32 localaddr, __u32 remoteaddr)
{
  return ssmsg_send_msg(localaddr, remoteaddr, SNAPSHOT_MSGTYPE_CLEAR,
                        0, NULL, 0);
}

int ssmsg_send_control(__u32 localaddr, __u32 remoteaddr, __u32 seqno)
{
  return ssmsg_send_msg(localaddr, remoteaddr, SNAPSHOT_MSGTYPE_CONTROL,
                        seqno, NULL, 0);
}

int ssmsg_send_marker(__u32 localaddr, __u32 remoteaddr, __u32 seqno)
{
  return ssmsg_send_msg(localaddr, remoteaddr, SNAPSHOT_MSGTYPE_MARKER,
                        seqno, NULL, 0);
}

int ssmsg_send_snapshot(__u32 localaddr, __u32 remoteaddr, __u32 seqno,
                        __u32 * ssdata, int ssdata_cnt)
{
  const int msghdr_len = sizeof(struct snapshot_msghdr_snapshot);
  struct snapshot_msghdr_snapshot *hdr;
  unsigned char *data;
  int data_len;
  int ret;

  data_len = msghdr_len + ssdata_cnt * sizeof(__u32);
  data = malloc(data_len);
  if (data == NULL)
    return -1;

  // SNAPSHOT message header
  hdr = (struct snapshot_msghdr_snapshot *) (data);
  hdr->entrycnt = htons(ssdata_cnt / 2);
  hdr->padding = 0;

  // SNAPSHOT message content
  memcpy(data + msghdr_len, ssdata, data_len - msghdr_len);

  ret = ssmsg_send_msg(localaddr, remoteaddr, SNAPSHOT_MSGTYPE_SNAPSHOT,
                       seqno, data, data_len);

  free(data);

  return ret;
}
