#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <linux/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <lib/util.h>
#include <router/ddos/snapshot/snapshot.h>
#include <router/ddos/snapshot/ssnode.h>

inline int ssnode_alloc_map(ssnode_map_t * map, int cnt)
{
  map->node = (ssnode_t *) malloc(cnt * sizeof(ssnode_t));
  if (map->node == NULL)
    return -1;
  map->cnt = cnt;
  return 0;
}

inline void ssnode_free_map(ssnode_map_t * map)
{
  if (map->node != NULL) {
    free(map->node);
    map->node = NULL;
  }
  map->cnt = 0;
}

inline int ssnode_alloc_link(ssnode_link_t * link, int cnt)
{
  link->nptr = (ssnode_t **) malloc(cnt * sizeof(ssnode_t *));
  if (link->nptr == NULL)
    return -1;

  link->channel_state = (dual_t *) malloc(cnt * sizeof(dual_t));
  if (link->channel_state == NULL) {
    free(link->nptr);
    return -1;
  }

  link->cnt = cnt;
  return 0;
}

inline void ssnode_free_link(ssnode_link_t * link)
{
  if (link->nptr != NULL) {
    free(link->nptr);
    link->nptr = NULL;
  }
  if (link->channel_state != NULL) {
    free(link->channel_state);
    link->channel_state = NULL;
  }
  link->cnt = 0;
}

inline int ssnode_node_index(ssnode_map_t * map, __u32 addr)
{
  int i;

  for (i = 0; i < map->cnt; i++) {
    if (map->node[i].addr == addr)
      return i;
  }

  return -1;
}

int ssnode_insert_links(ssnode_map_t * map, __u32 parent_addr,
                        __u32 * link_addr_map, int link_addr_cnt)
{
  int parent_idx, link_idx, i, k;
  ssnode_t *parent;

  if (link_addr_cnt == 0)
    return 0;

  parent_idx = ssnode_node_index(map, parent_addr);
  parent = &(map->node[parent_idx]);

  if (ssnode_alloc_link(&(parent->iup), link_addr_cnt) == -1)
    return -1;

  k = 0;
  for (i = 0; i < parent->iup.cnt; i++) {
    link_idx = ssnode_node_index(map, link_addr_map[i]);
    if (link_idx != -1) {
      parent->iup.nptr[k] = &(map->node[link_idx]);
      k++;
    }
  }

  return 0;
}

void ssnode_free_map_all(ssnode_map_t * map)
{
  int i;

  for (i = 0; i < map->cnt; i++) {
    ssnode_free_link(&(map->node[i].iup));
  }

  ssnode_free_map(map);
}

int ssnode_insert_nodes(ssnode_map_t * map, __u32 * node_addr_map,
                        int node_addr_cnt)
{
  int i;

  if (map->cnt != node_addr_cnt)
    return -1;

  for (i = 0; i < node_addr_cnt; i++) {
    map->node[i].addr = node_addr_map[i];
  }

  return 0;
}

void ssnode_print_dual_map(ssnode_map_t * map)
{
  struct in_addr in;
  ssnode_t *node;
  int i, j;
  int prev, curr;

  curr = map->dual_idx;
  prev = (map->dual_idx == 1) ? 0 : 1;

  printf("SNAPSHOT"
         "%22.22s %-18.18s  %-10.10s %-10.10s %-10.10s %-10.10s %-20.20s %-20.20s\n",
         " ", "C(H)", "UB", "L", "Ls", "LB", "DH", "N");

  for (i = 0; i < map->cnt; i++) {
    node = &(map->node[i]);
    in.s_addr = node->addr;
    printf
        ("%-16.16s        %8lld/%-8lld   %8lld %c %8lld %c %8lld %c %8lld   %8lld/%-8lld   %8lld/%-8lld\n",
         inet_ntoa(in), node->local_state.value[prev],
         node->local_state.value[curr], node->local_traffic_upper,
         (node->local_traffic_upper >=
          node->local_stat_traffic) ? ' ' : 'X', node->local_traffic,
         (node->local_traffic == node->local_stat_traffic) ? ' ' : 'X',
         node->local_stat_traffic,
         (node->local_traffic_lower <=
          node->local_stat_traffic) ? ' ' : 'X',
         node->local_traffic_lower,
         node->down_channel_state_sum.value[prev],
         node->down_channel_state_sum.value[curr],
         node->acc_local_traffic.value[prev],
         node->acc_local_traffic.value[curr]);

    if (node->iup.cnt != 0) {
      for (j = 0; j < node->iup.cnt; j++) {
        in.s_addr = node->iup.nptr[j]->addr;
        printf(" |__ %-16.16s   %8lld/%-8lld\n",
               inet_ntoa(in),
               node->iup.channel_state[j].value[prev],
               node->iup.channel_state[j].value[curr]);
      }
    }
  }
  printf("unmatched count = %d\n", map->unmatched_cnt);
}

void ssnode_print_map(ssnode_map_t * map, int dual_idx)
{
  struct in_addr in;
  ssnode_t *node;
  int i, j;

  printf("SNAPSHOT (%d) [%c]"
         "%12.12s %-10.10s  %-10.10s %-10.10s %-10.10s %-10.10s %-10.10s %-10.10s\n",
         dual_idx, (map->dual_idx == dual_idx) ? '*' : ' ',
         " ", "C(H)", "UB", "L", "Ls", "LB", "DH", "N");

  for (i = 0; i < map->cnt; i++) {
    node = &(map->node[i]);
    in.s_addr = node->addr;
    printf
        ("%-16.16s        %8lld   %8lld %c %8lld %c %8lld %c %8lld   %8lld   %8lld\n",
         inet_ntoa(in), node->local_state.value[dual_idx],
         node->local_traffic_upper,
         (node->local_traffic_upper >=
          node->local_stat_traffic) ? ' ' : 'X', node->local_traffic,
         (node->local_traffic == node->local_stat_traffic) ? ' ' : 'X',
         node->local_stat_traffic,
         (node->local_traffic_lower <=
          node->local_stat_traffic) ? ' ' : 'X',
         node->local_traffic_lower,
         node->down_channel_state_sum.value[dual_idx],
         node->acc_local_traffic.value[dual_idx]);

    if (node->iup.cnt != 0) {
      for (j = 0; j < node->iup.cnt; j++) {
        in.s_addr = node->iup.nptr[j]->addr;
        printf(" |__ %-16.16s   %8lld\n",
               inet_ntoa(in), node->iup.channel_state[j].value[dual_idx]);
      }
    }
  }
  printf("unmatched count = %d\n", map->unmatched_cnt);
}

void ssnode_print_cutoff_to_file(ssnode_map_t * map, double cutoff_value,
                                 const char *filename)
{
  FILE *fp;
  ssnode_t *node;
  int i, j, maxI;
  struct in_addr in;
  long long *data;
  long long total, sub, temp;
  unsigned int *addr, tempaddr;

  fp = fopen(filename, "w+");
  if (fp == NULL)
    return;

  data = malloc(map->cnt * sizeof(long long));
  if (data == NULL) {
    fclose(fp);
    return;
  }

  addr = malloc(map->cnt * sizeof(unsigned int));
  if (addr == NULL) {
    free(data);
    fclose(fp);
    return;
  }

  total = 0;
  for (i = 0; i < map->cnt; i++) {
    node = &(map->node[i]);
    data[i] = node->local_traffic;
    addr[i] = node->addr;
    total += data[i];
  }

  sub = 0;
  for (i = 0; i < map->cnt; i++) {
    maxI = i;
    for (j = i; j < map->cnt; j++) {
      if (data[j] > data[maxI])
        maxI = j;
    }

    in.s_addr = addr[maxI];
    fprintf(fp, "%s\n", inet_ntoa(in));

    temp = data[i];
    data[i] = data[maxI];
    data[maxI] = temp;

    tempaddr = addr[i];
    addr[i] = addr[maxI];
    addr[maxI] = tempaddr;

    sub += data[i];

    if (sub / (double) total >= cutoff_value)
      break;
  }

  free(data);
  free(addr);
  fclose(fp);
}

void ssnode_print_map_to_files(ssnode_map_t * map, int seq, FILE ** fp)
{
  ssnode_t *node;
  int i;

  for (i = 0; i < map->cnt; i++) {
    node = &(map->node[i]);
    fprintf(fp[i], "%d\t%8lld\t%8lld\t%8lld\t%8lld\n",
            seq,
            node->local_traffic_upper,
            node->local_traffic,
            node->local_stat_traffic, node->local_traffic_lower);
    fflush(fp[i]);
  }
}

void ssnode_analyse(ssnode_map_t * map, int dual_idx)
{
  ssnode_t *node;
  int i, j;
  long long victim_traffic_1 = 0;       // by local state of victim
  long long victim_traffic_2 = 0;       // by router_local_state_sum + deta^(1) - deta^(2)
  long long router_local_state_sum = 0;
  int prev, curr;

  // IMPORTANT NOTES !!!
  //
  // victim, all are its upstream routers
  // router, links[0] is its downstream router, others are upstream ones

  curr = map->dual_idx;
  prev = (map->dual_idx == 1) ? 0 : 1;

  map->acc_channel_state.value[dual_idx] = 0;

  for (i = 0; i < map->cnt; i++) {
    node = &(map->node[i]);

    // down_channel_state_sum (DH)
    if (i == 0)
      node->down_channel_state_sum.value[dual_idx] = 0;

    // acc_local_traffic (N)
    node->acc_local_traffic.value[dual_idx] =
        node->local_state.value[dual_idx];

    j = (i == 0) ? 0 : 1;
    for (; j < node->iup.cnt; j++) {

      // deta^(k)
      map->acc_channel_state.value[dual_idx] +=
          node->iup.channel_state[j].value[dual_idx];

      // down_channel_state_sum (DH)
      node->iup.nptr[j]->down_channel_state_sum.value[dual_idx] =
          node->down_channel_state_sum.value[dual_idx] +
          node->iup.channel_state[j].value[dual_idx];

      // acc_local_traffic (N)
      node->acc_local_traffic.value[dual_idx] -=
          node->iup.nptr[j]->local_state.value[dual_idx];
      node->acc_local_traffic.value[dual_idx] +=
          node->iup.channel_state[j].value[dual_idx];
    }

    // local_traffic (L), local_traffic_upper (UB), local_traffic_lower (LB)
    node->local_traffic =
        node->acc_local_traffic.value[curr] -
        node->acc_local_traffic.value[prev];
    node->local_traffic_upper =
        node->local_traffic + node->down_channel_state_sum.value[prev];
    node->local_traffic_lower =
        node->local_traffic - node->down_channel_state_sum.value[curr];

    if (i != 0) {
      router_local_state_sum += node->local_traffic;
    }
  }

  node = &(map->node[0]);
  usleep(10);
  victim_traffic_1 =
      node->local_state.value[curr] - node->local_state.value[prev];
  victim_traffic_2 =
      router_local_state_sum + map->acc_channel_state.value[prev] -
      map->acc_channel_state.value[curr];

  if (victim_traffic_1 != victim_traffic_2) {
    map->unmatched_cnt++;
  }

  printf("[%lld\t%lld] %s\n",
         victim_traffic_1,
         victim_traffic_2,
         (victim_traffic_1 == victim_traffic_2) ?
         "matched" : "!!!!!!!!!! not matched !!!!!!!!!!");
}

int ssnode_parse_snapshot(ssnode_map_t * map, unsigned char *data,
                          int datalen)
{
  static __u32 channel_state[MAX_LINK_LIMIT];
  struct snapshot_msghdr_snapshot *hdr =
      (struct snapshot_msghdr_snapshot *) data;
  __u32 *entry =
      (__u32 *) (data + sizeof(struct snapshot_msghdr_snapshot));
  __u16 entrycnt = ntohs(hdr->entrycnt);
  struct ssnode_t *node;
  int i, j, idx, cnt;
  int dual_idx = map->dual_idx;

  if ((datalen - sizeof(struct snapshot_msghdr_snapshot)) !=
      (entrycnt * sizeof(__u32) * 2)) {
    CERR_RET(-1, "data length error\n");
  }

  idx = ssnode_node_index(map, entry[0]);
  if (idx == -1) {
    CERR_RET(-1, "Invalid snapshot: no such node\n");
  }

  node = &(map->node[idx]);

  if (node->iup.cnt != (entrycnt - 1)) {
    CERR_RET(-1, "Invalid snapshot: number of entries not matched\n");
  }

  cnt = 0;

  for (i = 0; i < node->iup.cnt; i++) {
    for (j = 1; j < entrycnt; j++) {
      if (node->iup.nptr[i]->addr == entry[2 * j]) {
        channel_state[i] = ntohl(entry[2 * j + 1]);
        cnt++;
      }
    }
  }

  if (cnt != map->node[idx].iup.cnt) {
    CERR_RET(-1, "Invalid snapshot: missing entires\n");
  }

  for (i = 0; i < node->iup.cnt; i++) {
    node->iup.channel_state[i].value[dual_idx] = channel_state[i];
  }

  node->local_state.value[dual_idx] = ntohl(entry[1]);

  return 0;
}

int ssnode_parse_stat(ssnode_map_t * map, unsigned char *data, int datalen)
{
  static __u32 stat[MAX_STAT_LIMIT];
  struct snapshot_msghdr_stat *hdr = (struct snapshot_msghdr_stat *) data;
  __u32 *entry = (__u32 *) (data + sizeof(struct snapshot_msghdr_stat));
  __u16 entrycnt = ntohs(hdr->entrycnt);
  int i, j, cnt;

  if ((datalen - sizeof(struct snapshot_msghdr_stat)) !=
      (entrycnt * sizeof(__u32) * 2)) {
    CERR_RET(-1, "data length error\n");
  }

  if (entrycnt != map->cnt) {
    CERR_RET(-1, "Invalid snapshot: number of entries not matched\n");
  }

  cnt = 0;
  for (i = 0; i < map->cnt; i++) {
    for (j = 0; j < entrycnt; j++) {
      if (map->node[i].addr == entry[2 * j]) {
        stat[i] = ntohl(entry[2 * j + 1]);
        cnt++;
      }
    }
  }

  if (cnt != map->cnt) {
    CERR_RET(-1, "Invalid statistic: missing entires\n");
  }

  for (i = 0; i < map->cnt; i++) {
    map->node[i].local_stat_traffic = stat[i];
  }

  return 0;
}
