#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <lib/util.h>
#include <lib/net/rawsock.h>
#include <router/ddos/snapshot/snapshot.h>
#include <router/ddos/snapshot/ssmsg.h>
#include <router/ddos/snapshot/ssnode.h>
#include <router/ddos/snapshot/ssproc.h>
#include <signal.h>

#define __PRINT_MAP_TO_FILES__
//#define __INCLUDE_TEST__

#define DATA_BUFSIZE 2000
#define hline()                                               \
  printf("--------------------------------------------------" \
         "--------------------------------------------------" \
         "------------------------------\n")

#define NIPQUAD(addr) \
	((unsigned char *)&addr)[0], \
	((unsigned char *)&addr)[1], \
	((unsigned char *)&addr)[2], \
	((unsigned char *)&addr)[3]

#define DEBUGP_MSG(prefix, msgtype, seqno, iph)         \
  printf("%s %-15.15s (%u) from %u.%u.%u.%u\n", \
         prefix, msgtype, seqno, NIPQUAD(iph->saddr));

// program information
const char *program_version = "0.1";
const char *program_name = "ssproc";

static char snapshot_topology_file[] = "_snapshot.dat";

ssproc_t ssproc;
struct itimerval itv;

FILE **map_fp;
int map_idx;

double cutoff_value;

void *recv_thread_handle()
{
  rawsock_t raw;
  int totlen, datalen;
  struct iphdr *iph;
  struct snapshot_opt *opt;
  ssnode_map_t *map = &(ssproc.map);
//  addrlist_t *nodes = &(ssproc.nodes);
  struct timeval tv;
  unsigned long rtt;
  __u32 seqno;

  pthread_detach(pthread_self());
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  printf("entering recv loop ...\n");

  memset(&raw, 0, sizeof(rawsock_t));

  if (rawsock_alloc_data(&raw, DATA_BUFSIZE) == NULL) {
    ERR();
    goto end;
  }

  if (rawsock_create_socket(&raw, IPPROTO_SNAPSHOT) < 0) {
    ERR();
    goto end;
  }

  raw.opt.optlen = SNAPSHOT_OPTLEN;

  while ((totlen = rawsock_recv(&raw)) > 0) {
    ssproc_lock();

    iph = &(raw.iph);
    opt = (struct snapshot_opt *) &(raw.opt);
    seqno = ntohl(opt->seqno);
    datalen = totlen - SNAPSHOT_IPHLEN;

    switch (ssmsg_msgtype(&(raw.iph), opt)) {

    case SNAPSHOT_MSGTYPE_VMAP_ACK:
      DEBUGP_MSG("received", "VMAP ACK", seqno, iph);
      if ((ssproc.state == SSPROC_STATE_UNINITIALIZED) &&
          (seqno == ssproc.seqno)) {
        ssproc.map_unack_cnt--;
        if (ssproc.map_unack_cnt == 0) {
          ssproc.state = SSPROC_STATE_READY;
          printf("entered READY state\n");
        }
      }
      break;

    case SNAPSHOT_MSGTYPE_RMAP_ACK:
      DEBUGP_MSG("received", "RMAP ACK", seqno, iph);
      if ((ssproc.state == SSPROC_STATE_UNINITIALIZED) &&
          (seqno == ssproc.seqno)) {
        ssproc.map_unack_cnt--;
      }
      break;

    case SNAPSHOT_MSGTYPE_SMAP_ACK:
      DEBUGP_MSG("received", "SMAP ACK", seqno, iph);
      break;

    case SNAPSHOT_MSGTYPE_CONTROL_ACK:
      DEBUGP_MSG("received", "CONTROL ACK", seqno, iph);
      break;

    case SNAPSHOT_MSGTYPE_SNAPSHOT:
      DEBUGP_MSG("received", "SNAPSHOT", seqno, iph);

      if ((ssproc.state == SSPROC_STATE_STARTED) &&
          (seqno == ssproc.seqno)) {
        /*
           printf("totlen = %d\n", totlen);
           bufdump(iph, sizeof(struct iphdr));
           ssmsg_dump_iphdr(iph);
           hline();          
           bufdump(opt, SNAPSHOT_OPTLEN);
           ssmsg_dump_opt(opt);
           hline();
           printf("datalen = %d\n", datalen);
           bufdump(raw.data, datalen);
           ssmsg_dump_msg(opt, raw.data, datalen);
           hline();
         */
        if (ntohl(opt->seqno) != ssproc.seqno) {
          printf("snapshot seqno not matched (opt:%u  ssproc:%u)\n",
                 ntohl(opt->seqno), ssproc.seqno);
        } else {
          ssnode_parse_snapshot(map, raw.data, datalen);
          map->snapshotcnt++;

          if (map->snapshotcnt == map->cnt) {
            /*
               gettimeofday(&tv, NULL);
               rtt = ((tv.tv_sec - ssproc.tv.tv_sec) * 1000000
               + (tv.tv_usec - ssproc.tv.tv_usec));
               printf("RTT = %lu (us)\n", rtt);
             */
            ssproc.state = SSPROC_STATE_FINISHED;
            ssnode_analyse(map, map->dual_idx);

            if (map_idx > 0) {
#ifdef __PRINT_MAP_TO_FILES__
              ssnode_print_map_to_files(map, map_idx, map_fp);
              ssnode_print_cutoff_to_file(map, cutoff_value,
                                          "_cutoff.dat");
#endif
              ssnode_print_dual_map(map);
              hline();
            }
            // ssnode_print_map(map, map->dual_idx); hline();
            map->dual_idx = (map->dual_idx == 0) ? 1 : 0;
            map->snapshotcnt = 0;
            map_idx++;
          }
        }
      }
      break;

    case SNAPSHOT_MSGTYPE_STAT:
      DEBUGP_MSG("received", "STAT", seqno, iph);
      if ((ssproc.state == SSPROC_STATE_STARTED) &&
          (seqno == ssproc.seqno)) {
        //ssmsg_dump_msg(opt, raw.data, datalen);
        ssnode_parse_stat(map, raw.data, datalen);
      }
      break;
    }

    if ((ssproc.state == SSPROC_STATE_UNINITIALIZED) &&
        (ssproc.map_unack_cnt == 0)) {
      ssproc.state = SSPROC_STATE_READY;
      printf("entered READY state\n");
      setitimer(ITIMER_REAL, &itv, NULL);
    }

    ssproc_unlock();
  }

  rawsock_free_data(&raw);

  close(raw.sd);

end:
  pthread_exit(NULL);
  return NULL;
}

// -- FOR TESTING PURPOSE ONLY --------------------------------------------------
#ifdef __INCLUDE_TEST__

#define MAP_LINK(i, link) {                    \
  links[i].cnt = sizeof(link) / sizeof(__u32); \
  links[i].addr = link; \
}

#define LIST_CNT(list) (sizeof(list)/sizeof(__u32))
#define HTONL_LIST(i, list) {      \
  for (i=0; i<LIST_CNT(list); i++) \
    list[i] = htonl(list[i]);      \
}

int build_test_links()
{
  static __u32 node[] = {
    0xc80100fe, 0xc8010101, 0xc8010202, 0xc8010303, 0xc8010404, 0xc8010505,
    0xc8010606, 0xc8010707, 0xc8010808, 0xc8010909, 0xc8010a0a, 0xc8010b0b
  };
  static __u32 link0[] = { 0xc8010101, 0xc8010808 };
  static __u32 link1[] = { 0xc80100fe, 0xc8010202, 0xc8010707 };
  static __u32 link2[] = { 0xc8010101, 0xc8010303 };
  static __u32 link3[] = { 0xc8010202, 0xc8010404 };
  static __u32 link4[] = { 0xc8010303, 0xc8010505 };
  static __u32 link5[] = { 0xc8010404, 0xc8010606 };
  static __u32 link6[] = { 0xc8010505 };
  static __u32 link7[] = { 0xc8010101 };
  static __u32 link8[] =
      { 0xc80100fe, 0xc8010909, 0xc8010a0a, 0xc8010b0b };
  static __u32 link9[] = { 0xc8010808 };
  static __u32 link10[] = { 0xc8010808 };
  static __u32 link11[] = { 0xc8010808 };

  addrlist_t *nodes = &(ssproc.nodes);
  addrlist_t *links = NULL;
  int i;

  printf("building test links ...\n");

  HTONL_LIST(i, node);
  HTONL_LIST(i, link0);
  HTONL_LIST(i, link1);
  HTONL_LIST(i, link2);
  HTONL_LIST(i, link3);
  HTONL_LIST(i, link4);
  HTONL_LIST(i, link5);
  HTONL_LIST(i, link6);
  HTONL_LIST(i, link7);
  HTONL_LIST(i, link8);
  HTONL_LIST(i, link9);
  HTONL_LIST(i, link10);
  HTONL_LIST(i, link11);

  nodes->cnt = LIST_CNT(node);
  nodes->addr = node;

  links = (addrlist_t *) malloc(nodes->cnt * sizeof(addrlist_t));
  if (links == NULL)
    return -1;

  // victim
  MAP_LINK(0, link0);

  // router
  MAP_LINK(1, link1);
  MAP_LINK(2, link2);
  MAP_LINK(3, link3);
  MAP_LINK(4, link4);
  MAP_LINK(5, link5);
  MAP_LINK(6, link6);
  MAP_LINK(7, link7);
  MAP_LINK(8, link8);
  MAP_LINK(9, link9);
  MAP_LINK(10, link10);
  MAP_LINK(11, link11);

  ssproc.links = links;

  return 0;
}

#endif
// ------------------------------------------------------------------------------

int build_links()
{
  addrlist_t *nodes = &(ssproc.nodes);
  addrlist_t *links = NULL;
  addrlist_t *link = NULL;
  int i;
  FILE *fp = NULL;
  unsigned long data;

  printf("building links ...\n");
  fp = fopen(snapshot_topology_file, "r");
  if (fp == NULL) {
    CERR("Fail to read topology data file \"%s\".\n", snapshot_topology_file);
    return -1;
  }
  // check file size !!!

  nodes->cnt = 0;
  nodes->addr = NULL;
  do {
    fread(&data, sizeof(data), 1, fp);
    if (data != 0xFFFFFFFF) {
      nodes->cnt++;
      if (nodes->addr == NULL) {
        nodes->addr = malloc(sizeof(data));
      } else {
        nodes->addr = realloc(nodes->addr, nodes->cnt * sizeof(data));
      }
      nodes->addr[nodes->cnt - 1] = data;
    }
  } while (data != 0xFFFFFFFF);

  links = (addrlist_t *) malloc(nodes->cnt * sizeof(addrlist_t));
  if (links == NULL)
    return -1;

  for (i = 0; i < nodes->cnt; i++) {
    link = &(links[i]);
    link->cnt = 0;
    link->addr = NULL;
    do {
      fread(&data, sizeof(data), 1, fp);
      if (data != 0xFFFFFFFF) {
        link->cnt++;
        if (link->addr == NULL) {
          link->addr = malloc(sizeof(data));
        } else {
          link->addr = realloc(link->addr, link->cnt * sizeof(data));
        }
        link->addr[link->cnt - 1] = data;
      }
    } while (data != 0xFFFFFFFF);
  }
  ssproc.links = links;

  return 0;
}

void free_links()
{
  addrlist_t *nodes = &(ssproc.nodes);
  addrlist_t *links = ssproc.links;
  int i;

  if (nodes->addr != NULL) {
    free(nodes->addr);
    nodes->addr = NULL;
  }
  for (i = 0; i < nodes->cnt; i++) {
    if (links[i].addr != NULL) {
      free(links[i].addr);
      links[i].addr = NULL;
    }
  }
}

int open_map_files()
{
  struct in_addr in;
  addrlist_t *nodes = &(ssproc.nodes);
  static char temp[1000];
  int i, j;

  map_fp = malloc(nodes->cnt * sizeof(FILE *));
  if (map_fp == NULL)
    goto end_err;

  for (i = 0; i < nodes->cnt; i++) {
    in.s_addr = nodes->addr[i];
    snprintf(temp, sizeof(temp) - 1, "_routerData_%s", inet_ntoa(in));
    map_fp[i] = fopen(temp, "w+");
    if (map_fp[i] == NULL) {
      for (j = 0; j < i; j++) {
        fclose(map_fp[j]);
      }
      goto end_free_map_fp;
    }
  }

  return 0;

end_free_map_fp:
  free(map_fp);

end_err:
  return -1;
}

void close_map_files()
{
  addrlist_t *nodes = &(ssproc.nodes);
  int i;

  if (map_fp == NULL)
    return;

  for (i = 0; i < nodes->cnt; i++) {
    if (map_fp[i] != NULL) {
      fclose(map_fp[i]);
    }
  }
  free(map_fp);
}

int build_map()
{
  ssnode_map_t *map = &(ssproc.map);
  addrlist_t *nodes = &(ssproc.nodes);
  addrlist_t *links = ssproc.links;
  int i;

  printf("building map ...\n");

  memset(map, 0, sizeof(ssnode_map_t));
  if (ssnode_alloc_map(map, nodes->cnt) == -1)
    return -1;

  memset(map->node, 0, map->cnt * sizeof(ssnode_t));
  ssnode_insert_nodes(map, nodes->addr, nodes->cnt);

  for (i = 0; i < nodes->cnt; i++) {
    if ((links[i].cnt != 0) && (links[i].addr != NULL)) {
      ssnode_insert_links(map, nodes->addr[i], links[i].addr,
                          links[i].cnt);
    }
  }

  map->snapshotcnt = 0;
  map->dual_idx = 0;

  return 0;
}

int send_reset()
{
  addrlist_t *nodes = &(ssproc.nodes);
  int i, rret = 0, ret = 0;

  printf("resetting participating routers ...\n");

  for (i = 0; i < nodes->cnt; i++) {
    printf("   RESET %u.%u.%u.%u -> %u.%u.%u.%u ... ",
           NIPQUAD(nodes->addr[0]), NIPQUAD(nodes->addr[i]));
    ret = ssmsg_send_reset(nodes->addr[0], nodes->addr[i]);
    if (ret != 0) {
      printf("FAILED\n");
      rret = ret;
    } else {
      printf("OK\n");
    }
  }

  return rret;
}

int send_clear()
{
  addrlist_t *nodes = &(ssproc.nodes);
  int i, rret = 0, ret = 0;

  printf("clearing participating routers ...\n");

  for (i = 0; i < nodes->cnt; i++) {
    printf("   CLEAR %u.%u.%u.%u -> %u.%u.%u.%u ... ",
           NIPQUAD(nodes->addr[0]), NIPQUAD(nodes->addr[i]));
    ret = ssmsg_send_clear(nodes->addr[0], nodes->addr[i]);
    if (ret != 0) {
      printf("FAILED\n");
      rret = ret;
    } else {
      printf("OK\n");
    }
  }

  return rret;
}

int send_map()
{
  addrlist_t *nodes = &(ssproc.nodes);
  addrlist_t *links = ssproc.links;

  __u32 tmp[100];
  int i, ret = 0;

  printf("sending map to participating routers ...\n");

  printf("    SMAP %u.%u.%u.%u -> %u.%u.%u.%u\n",
         NIPQUAD(nodes->addr[0]), NIPQUAD(nodes->addr[0]));
  ret = ssmsg_send_smap(nodes->addr[0], nodes->addr[0],
                        nodes->addr, nodes->cnt);

  printf("    VMAP %u.%u.%u.%u -> %u.%u.%u.%u\n",
         NIPQUAD(nodes->addr[0]), NIPQUAD(nodes->addr[0]));
  ret = ssmsg_send_vmap(nodes->addr[0], nodes->addr[0], nodes->cnt - 1,
                        links[0].addr, links[0].cnt);
  if (ret != 0)
    goto end;
  ssproc.map_unack_cnt = 1;

  for (i = 1; i < nodes->cnt; i++) {
    if (links[i].addr[0] == nodes->addr[0]) {
      printf("    RMAP %u.%u.%u.%u -> %u.%u.%u.%u ... ",
             NIPQUAD(nodes->addr[0]), NIPQUAD(nodes->addr[i]));
      ret = ssmsg_send_rmap(nodes->addr[0], nodes->addr[i], 1,
                            links[i].addr, links[i].cnt);
      if (ret != 0) {
        printf("FAILED\n");
        goto end;
      } else {
        printf("OK\n");
      }
      ssproc.map_unack_cnt++;
    } else {
      memcpy(&(tmp[1]), links[i].addr, links[i].cnt * sizeof(__u32));
      tmp[0] = nodes->addr[0];
      printf("    RMAP %u.%u.%u.%u -> %u.%u.%u.%u ... ",
             NIPQUAD(nodes->addr[0]), NIPQUAD(nodes->addr[i]));
      ret = ssmsg_send_rmap(nodes->addr[0], nodes->addr[i], 0,
                            tmp, links[i].cnt + 1);
      if (ret != 0) {
        printf("FAILED\n");
        goto end;
      } else {
        printf("OK\n");
      }
      ssproc.map_unack_cnt++;
    }
  }

end:
  return ret;
}

void timer_event(int signum)
{
  addrlist_t *nodes = &(ssproc.nodes);
  ssproc_lock();
  printf("timer event\n");
  if ((ssproc.state == SSPROC_STATE_READY) ||
      (ssproc.state == SSPROC_STATE_FINISHED)) {
    gettimeofday(&(ssproc.tv), NULL);
    ssproc.seqno = ssproc.tv.tv_sec * 1000000 + ssproc.tv.tv_usec;
    printf("sending CONTROL to %u.%u.%u.%u ... (seqno = %u)\n",
           NIPQUAD(nodes->addr[0]), ssproc.seqno);
    ssmsg_send_control(nodes->addr[0], nodes->addr[0], ssproc.seqno);
    ssproc.state = SSPROC_STATE_STARTED;
    printf("entered STARTED state\n");
  }
  ssproc_unlock();
}

void cleanup()
{
  signal(SIGALRM, SIG_DFL);
  pthread_cancel(ssproc.recv_thread);

#ifdef __PRINT_MAP_TO_FILES__
  close_map_files();
#endif

  ssproc_unlock();
  ssnode_free_map_all(&(ssproc.map));

  free_links();
  free(ssproc.links);

  pthread_mutex_destroy(&(ssproc.mutex));
}

static void signal_handler(int signo)
{
  switch (signo) {
  case SIGQUIT:
  case SIGINT:
  case SIGHUP:
    cleanup();
    exit(0);
    break;
  }
}

static void catch_signal()
{
  signal(SIGQUIT, signal_handler);
  signal(SIGINT, signal_handler);
  signal(SIGHUP, signal_handler);
}

int main(int argc, char **argv)
{
  int ret = 0;
  int resetflag = 0;

  if (argc != 3) {
    printf("%s v%s\n\n"
           "Usage: %s reset cutoff\n\n"
           "Arguments:\n"
           "    reset  - \"1\" to reset all snapshot clients\n"
           "    cutoff - cutoff percentage [0..1]\n\n",
           program_name, program_version, program_name);
    exit(1);
  }

  if (atoi(argv[1]) == 1) {
    resetflag = 1;
  }

  cutoff_value = atof(argv[2]);

  itv.it_interval.tv_sec = 0;
  itv.it_interval.tv_usec = 1000000;
  itv.it_value = itv.it_interval;

  ssproc.state = SSPROC_STATE_UNINITIALIZED;
  ssproc.map_unack_cnt = 0;
  pthread_mutex_init(&(ssproc.mutex), NULL);

  if (build_links() != 0) {
    ret = 1;
    goto end;
  }

  if (build_map() != 0) {
    ret = 1;
    goto end_free_links;
  }
#ifdef __PRINT_MAP_TO_FILES__
  open_map_files();
  map_idx = 0;
#endif

  pthread_create(&(ssproc.recv_thread), NULL, recv_thread_handle, NULL);

  signal(SIGALRM, timer_event);

  ssproc_lock();
  if (resetflag == 1) {
    if (send_reset() != 0)
      goto end_free_map;

    sleep(1);

    if (send_map() != 0)
      goto end_free_map;
  } else {
    if (send_clear() != 0)
      goto end_free_map;

    ssproc.state = SSPROC_STATE_READY;
    printf("entered READY state\n");
    setitimer(ITIMER_REAL, &itv, NULL);
  }
  ssproc_unlock();

  catch_signal();

  while (1) {
    sleep(10);
  }

  signal(SIGALRM, SIG_DFL);
  pthread_cancel(ssproc.recv_thread);

#ifdef __PRINT_MAP_TO_FILES__
  close_map_files();
#endif

end_free_map:
  ssproc_unlock();
  ssnode_free_map_all(&(ssproc.map));

end_free_links:
  free_links();
  free(ssproc.links);

end:
  pthread_mutex_destroy(&(ssproc.mutex));
  return ret;
}
