#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <router/ddos/traceback/traceback.h>

#define MAX_CNT 10000

void printMarkingFieldsNodeList2(const MarkingFieldsNode* list)
{
  const MarkingFieldsNode* ptr;

  if (list == NULL)
    return;
  
  for (ptr=list->next; ptr!=NULL; ptr=ptr->next)
  {
    printf("%d\t", ptr->frequency);
  }
}

int printGraphList2(Router* head)
{
  Router* ptr;
  Router* ptr2;
  struct in_addr in;
  
  if (head == NULL)
  {
    printf("(NULL)\n");
    return 0;
  }
  ptr = head->nextRouter;
  while (1)
  {
    if (ptr == NULL)
      break;
//    in.s_addr = htonl(ptr->ip);
//    printf("%s(%lf) -> ", inet_ntoa(in), ptr->localTrafficRate);
    ptr2 = ptr->nextConnection;
    while (1)
    {
      if (ptr2 == NULL)
        break;
      in.s_addr = htonl(ptr2->ip);
      printf("%lf\t", ptr2->localTrafficRate);
      ptr2 = ptr2->nextConnection;
    }
    ptr = ptr->nextRouter;
  }
  return 1;
}

void printMarkingFields(MarkingFields *markingFields)
{
  static char startip[17], endip[17];
  struct in_addr in;

  in.s_addr = markingFields->startIP;
  strncpy(startip, inet_ntoa(in), sizeof(startip)-1);
  startip[sizeof(startip)-1] = '\0';
  
  in.s_addr = markingFields->endIP;
  strncpy(endip, inet_ntoa(in), sizeof(endip)-1);
  startip[sizeof(endip)-1] = '\0';
  
  printf("%16s\t->\t%16s\t%3d\n",
         startip, endip, ntohl(markingFields->distance));
}

void construct(double markprob, double accuracy,
               MarkingFieldsNode* fieldList, MarkingFields *fields,
               int cnt, unsigned long elapsed)
{
  Router* routerHead;
  RouterTree* head = newRouterTree(0, 10);
  int i;

  printf("%lf\t", elapsed / 1000000.0);

  //  1. add all the fields into a MarkingFieldNode structure
  for (i=0; i<cnt; i++) {
    addMarkingFields(fieldList, &fields[i]);
  }
  printMarkingFieldsNodeList2(fieldList);
                
  //  2. Construct attack Graph, stored in a RouterTree structure
  attackGraphCon(head, fieldList);
//  printTree(head, NULL);

  //  3. Calculate the local traffic rate of every node
  calLocalTrafficRate(head, fieldList, elapsed, markprob);
//  printf("calLocalTrafficRate success\n");
//  printTree(head, NULL);
  
  //  4. Construct attack paths, stored in a Router structure
  routerHead = attackPathCon(head);
  printGraphList2(routerHead);

/*
//  5. For all attack paths...
for (routerPtr=routerHead->nextRouter; routerPtr!=NULL; 
routerPtr= routerPtr->nextRouter)
{
double localTR[10];
Router *rPtr;
for (rPtr=routerPtr->nextConnection, i=0; rPtr!=NULL;
rPtr=rPtr->nextConnection, i++)
{
localTR[i] = rPtr->localTrafficRate;
printf("localTrafficRate: %lf\n", localTR[i]);
}    

//  6. Calculate the minimum stable time
mst = calMinStableTime(markprob, theoretical_mst, 13.0, i, localTR);
printf("mTime: %lf\n", mst);
}
*/
   freeList(routerHead);
   free(head);
   printf("\n");
}

int main(int argc, char **argv)
{
  double markprob, accuracy, theoretical_mst;
  MarkingFields fields[MAX_CNT];
  MarkingFieldsNode* fieldList = NULL;
  unsigned long elapsed = 0;
  double t;
  int cnt = 0;
  int ret, state;

  int div;
  
  if (argc != 4) {
    printf("Usage: %s <marking probability> <accuracy> <minimum stable time>\n",
           argv[0]);
    exit(1);
  }
  
  markprob = atof(argv[1]);
  accuracy = atof(argv[2]);
  theoretical_mst = atof(argv[3]);

  /*
    printf("marking probability = %lf\n", markprob);
    printf("accuracy            = %lf\n", accuracy);
    printf("minimum stable time = %lf\n", theoretical_mst);
  */
  
  fieldList = newMarkingFieldsNode();

  state = 1;
  do {
    ret = scanf("%lu\t%lu\t%lu\t%lu\n",
                &elapsed,
                &fields[cnt].startIP,
                &fields[cnt].endIP,         
                &fields[cnt].distance);

    if (ret == EOF)
      break;

    /*
      printf("%lu\t%lu\t%lu\t%lu\n",
      elapsed,
      fields[cnt].startIP,
      fields[cnt].endIP,         
      fields[cnt].distance);
    */
    cnt++;

    t = elapsed/1000000.0;

    div = 4;
    if (t > state * theoretical_mst / div) {
      printf("%.2lf\t", ((double) state) / div);
      construct(markprob, accuracy, fieldList, fields, cnt,
                elapsed);
      state++;
      cnt = 0;
    } else {
      if (cnt == MAX_CNT) {
        construct(markprob, accuracy, fieldList, fields, cnt,
                  elapsed);
        cnt = 0;
      }
    }
  } while (state <= 10 * div);

  return 0;

}
