#include <assert.h>
#include <router/ddos/traceback/traceback.h>

////////////////////////////////////////////////////////

double calTrafficRateIn(const MarkingFieldsNode* fieldList, const double* trafficRate,
                        unsigned long toIp)
{
  int i;
  double tRateIn = 0;
  MarkingFieldsNode* ptr;
    
  for (ptr = fieldList->next, i=0; ptr!=NULL; ptr=ptr->next, i++)
  {
    if (ptr->markingField.endIP == toIp)
      tRateIn += trafficRate[i];
  }
  return tRateIn;
}

double calTrafficRateOut(const MarkingFieldsNode* fieldList, const double* trafficRate,
                         unsigned long fromIp)
{
  int i;
  double tRateOut = 0;
  MarkingFieldsNode* ptr;
    
  for (ptr = fieldList->next, i=0; ptr!=NULL; ptr=ptr->next, i++)
  {
    if (ptr->markingField.startIP == fromIp)
      tRateOut += trafficRate[i];
  }
  return tRateOut;
}


int numOfNode(const MarkingFieldsNode* fieldList)
{
  int i=0;
  MarkingFieldsNode* ptr;
    
  for (ptr = fieldList->next; ptr!=NULL; ptr=ptr->next)
  {
    i++;
  }
  return i;
}

int calLocalTrafficRateNode(RouterTree* node, const MarkingFieldsNode* fieldList,
                            const double* trafficRate)
{
  int i;
  /*
   *  recusion to the leaf nodes
   */
  for (i=0; i<node->numOfBranch; i++)
  {
    calLocalTrafficRateNode(node->nextLevelArray[i], fieldList, trafficRate);
  }

  // if this is leaf node ...
  if (node->numOfBranch == 0)
  {
    node->localTrafficRate = calTrafficRateOut(fieldList, trafficRate, node->ip);
  }
  else    // else if this is not leaf node ...
  {
    node->localTrafficRate = calTrafficRateOut(fieldList, trafficRate, node->ip) -
      calTrafficRateIn(fieldList, trafficRate, node->ip);
  }
  return 1;
}


/*
 *  calLocalTrafficRate
 *
 *  The tRate array store the traffic rates,
 *  each entry corresponse to the fieldList entry.
 *
 */
int calLocalTrafficRate(RouterTree* head, const MarkingFieldsNode* fieldList,
                        int time, double markp)
{
  double* tRate;
  MarkingFieldsNode* ptr;
  int i;
  int _numOfNode;

  assert(time != 0);
  assert(markp != 0);
  _numOfNode = numOfNode(fieldList);
  assert(_numOfNode != 0);
  tRate = (double*)malloc(_numOfNode * sizeof(double));
    
  /*
   *  Figure 10 in the paper 
   *  (e.distance-1 = "distance" field in the marking field
   */
  for (ptr = fieldList->next, i=0; ptr!=NULL; ptr=ptr->next, i++)
  {
    tRate[i] = ptr->frequency / (time * markp * 
                                 pow((1-markp), ptr->markingField.distance));
  }
  calLocalTrafficRateNode(head, fieldList, tRate);
    
  free(tRate);
  return 1;
}
