// indicate this is an ipqext extension
#define _IPQEXT_EXTENSION_

#define __DEBUG__

//----------------------------------------------------------------------//
//
//  ipqext extension module
//
//----------------------------------------------------------------------//


//----------------------------------------------------------------------//
//
//  REMARKS
//
//  single flow to multiple flow
//
//----------------------------------------------------------------------//

#include <asm/types.h>
#include <netinet/in.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <libipq.h>
#include <lib/util.h>
#include <router/ipqext/rsp.h>
#include <router/ipqext/ipqext.h>
#include <router/ipqext/extensions/ipqext_fec/rawsock.h>
#include <router/ipqext/extensions/ipqext_fec/decqueue.h>
#include <router/ipqext/extensions/ipqext_fec/encqueue.h>

/*
 * define extension name here
 *
 * extension name must match with the output module name
 * ( i.e. ipqext_{extension name} )
 */
const char extension_name[] = "fec";

//unsigned long counter = 0;
unsigned int in_seqno = 0;

int param_ksize;
int param_nsize;
int param_bsize;

//----------------------------------------------------------------------//
//
//  FEC encode process
//

eq_queue_t fec_out_queue;

static void process_udppkt(struct pkt_layout *pl)
{
  struct raw_struct raw;
  int fec_ret = eq_enqueue_buffer(&fec_out_queue, pl->upph);
  int i;

  switch (fec_ret) {
  case -1:
    CERR("eq_queue_add: error\n");
  case 1:
    eq_encode_queue(&fec_out_queue);
    eq_add_fechdr(&fec_out_queue, in_seqno);

    memset(&raw, 0, sizeof(struct raw_struct));

    if (create_raw_socket(&raw) < 0) {
      CERR("ERROR: opening raw socket\n");
    } else {
      raw.data_len = fec_out_queue.totsize;
      raw.src_addr.s_addr = pl->iph->saddr;
      raw.dst_addr.s_addr = pl->iph->daddr;
      raw.protocol = IPPROTO_RSP;
      raw.rsph.sid = RSPSERV_FEC;

      for (i=0; i<fec_out_queue.n; i++) {
        raw.data = eq_get_slot_head(fec_out_queue.buf[i]);
        send_raw_packet(&raw);
      }
      close(raw.sd);      
    }
    eq_flush_queue(&fec_out_queue);
    in_seqno++;
  default:
    break;
  }
}


static inline void init_encode()
{
  DB("initializing encode ...\n");
  memset(&fec_out_queue, 0, sizeof(eq_queue_t));
  //  eq_init_queue(&fec_out_queue, 8, 16, 1032);
  eq_init_queue(&fec_out_queue, param_ksize, param_nsize, param_bsize);
}

static inline void fini_encode()
{
  DB("finializing encode ...\n");
  eq_fini_queue(&fec_out_queue);
}


//----------------------------------------------------------------------//
//
//  FEC decode process
//

pthread_t decode_thread;
dq_queue_t *decode_queue;
__u32	decode_saddr;
__u32	decode_daddr;

static void process_fecpkt(struct pkt_layout *pl)
{
  dq_block_t *blk;
  dq_buffer_t *buf;
  struct fechdr *fech;
  int ret;

  fech = (struct fechdr *) pl->data;
  fech->seqno = ntohl(fech->seqno);
  fech->size = ntohs(fech->size);

  ret = 0;

  // check IP address
  if ((pl->iph->saddr != decode_queue->src_addr) ||
      (pl->iph->daddr != decode_queue->dst_addr)) {
    CERR("address mismatch: (%x,%x) != (%x,%x)\n",
         pl->iph->saddr, pl->iph->daddr,
         decode_queue->src_addr, decode_queue->dst_addr);
    return;
  }

  // check FEC header consistency
  if ((fech->k != decode_queue->k_size) || (fech->n != decode_queue->n_size)) {
    CERR("inconsistency (k,n): %d %d\n", fech->k, fech->n);
    return;
  }

  blk = dq_get_block(decode_queue, fech->seqno);
  if (blk == NULL) {
    // first block
    ret = dq_enqueue_create_block(decode_queue, &blk, fech->seqno, fech->size);
    if (ret != 0) return;
  } else {
    if (fech->size != blk->b_size) {
      CERR("inconsistency block size: %d\n", fech->size);
      ret = -1;
    }
  }
  if (ret != 0) return;
  ret = dq_new_copy_buffer(&buf, pl->data + sizeof(struct fechdr),
                           fech->size);
  if (ret != 0) return;

  dq_add_buffer(blk, fech->index, buf);
}

static void *decode()
{
  dq_block_t *blk;
  dq_decinfo_t *dec;
  struct raw_struct raw;
  int i;

  pthread_detach(pthread_self());
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  while (1) {
    blk = dq_dequeue_block(decode_queue);
    
    if (blk != NULL) {
      if (dq_new_decinfo(&dec, blk->k_size) == 0) {
        if (dq_decode_block(blk, dec, decode_queue->fec) != -1) {
          if (create_raw_socket(&raw) < 0) {
            CERR("ERROR: opening raw socket\n");
          } else {        
            raw.data_len = blk->b_size;
            raw.src_addr.s_addr = decode_queue->src_addr;
            raw.dst_addr.s_addr = decode_queue->dst_addr;
            raw.protocol = IPPROTO_UDP;
            for (i=0; i<blk->k_size; i++) {
              if (dec->buf[i] != NULL) {
                raw.data = dec->buf[i];
                send_raw_packet(&raw);
              }
            }
            close(raw.sd);
          }
          dq_free_decinfo(&dec);
        } else CERR("decode fail\n");
        dq_free_block(&blk);
      } else CERR("decinfo fail\n");
    }
  }

  pthread_exit(NULL);
  return NULL;
}

static inline int init_decode()
{
  int retval = 0;

  DB("initializing decode ...\n");

  //  if (dq_new_queue(&decode_queue, 0, 1000, 8, 16, 0x32f9a8c0, 0x33f9a8c0) != 0) {
  if (dq_new_queue(&decode_queue, 0, 1000, param_ksize, param_nsize, 0x32f9a8c0, 0x33f9a8c0) != 0) {
    return -1;
  }

  if (pthread_create(&decode_thread, NULL, decode, NULL) != 0) {
    perror(__FUNCTION__);

    dq_free_queue(&decode_queue);
    return -1;
  }

  return retval;
}

static inline void fini_decode()
{
  DB("finializing decode ...\n");
  pthread_cancel(decode_thread);
  dq_free_queue(&decode_queue);
}


//----------------------------------------------------------------------//
//
//  module functions
//

/*
 * ipqext structure
 */
static ipqext_t ext = {
  prev      : NULL,
  next      : NULL,
  handle    : NULL,
  active    : 0,
  name      : extension_name,
  version   : "1.0",
  groupmask : 0,
  init      : &init,
  fini      : &fini,
  proc      : &proc,
  start     : &start,
  stop      : &stop,
  reset     : &reset,
  info      : &info,
  help      : &help,
};

/*
 * packet processing function
 *
 * return: NF_ACCEPT   packet is accepted and leaves ipqext immediately
 *         NF_QUEUE    packet is queued for further ipqext processing
 *         NF_DROP     packet is explicitly dropped
 *
 */
int proc(ipq_packet_msg_t *m)
{
  struct pkt_layout pl;
  unsigned int retval = NF_QUEUE;
  struct rsphdr *rsph;

  pl.iph = (struct iphdr *) m->payload;
  pl.iph_len = pl.iph->ihl << 2;
  pl.upph = (((char *) pl.iph) + pl.iph_len);

  switch (pl.iph->protocol) {

  case IPPROTO_UDP:
    pl.data = (u_char *) (((char *) pl.upph) + sizeof(struct udphdr));
    pl.data_len = m->data_len - pl.iph_len - sizeof(struct udphdr);
    process_udppkt(&pl);
    return NF_DROP;

  case IPPROTO_RSP:
    pl.data = (u_char *) (((char *) pl.upph) + sizeof(struct rsphdr));
    pl.data_len = m->data_len - pl.iph_len - sizeof(struct rsphdr);
    rsph = (struct rsphdr *) pl.upph;
    if (rsph->sid == RSPSERV_FEC)
      process_fecpkt(&pl);
    return NF_DROP;
  }

  return retval;
}

/*
 * extension start function
 *
 * return on success, -1 or nonzero value on error
 */
int start()
{
  init_encode();
  init_decode();
  return 0;
}

/*
 * extension stop function
 *
 * return on success, -1 or nonzero value on error
 */
int stop()
{
  fini_decode();
  fini_encode();
  return 0;
}

/*
 * extension reset function
 *
 * return on success, -1 or nonzero value on error
 */
int reset()
{
  in_seqno = 0;
  decode_queue->headid = 0;
  return 0;
}

/*
 * extension info function
 *
 * return a pointer to a char buffer or NULL
 */
const char infobuf[] = " info here ";

const char *info()
{
  return infobuf;
}

/*
 * extension help function
 *
 * return a pointer to a char buffer or NULL
 */
const char helpbuf[] = " help here ";

const char *help()
{
  DB("%s: help invoked\n", extension_name);
  return helpbuf;
}

/*
 *  initial pass function
 *
 * return on success, -1 or nonzero value on error
 */
int init_pass(void *handle, char *param, int paramlen)
{
  char temp[IPQEXT_PARAM_LEN];
  int i, st, ed, idx, val;

  // set module handle
  ext.handle = handle;

  // parse parameter buffer
  if ((param == NULL) || (paramlen == 0))
    return -1;

  st = 0;
  idx = 0;

  for (i=0; i<paramlen; i++) {
    if ((param[i] == ',') || (param[i] == '\0')) {
      ed = i;

      LOG(LOG_INFO, "%d %d %d\n", st, ed, idx);

      if (ed-st > 0) {
        strncpy(temp, &(param[st]), ed-st);
        temp[ed-st] = '\0';

        LOG(LOG_INFO, "%s\n", temp);

        val = atoi(temp);
        if (val == 0) return -1;

        switch (idx) {
        case 0:
          param_ksize = val;
          idx++;
          break;

        case 1:
          param_nsize = val;
          idx++;

          if (param_nsize <= param_ksize)
            return -1;
          break;

        case 2:
          param_bsize = val;
          LOG(LOG_INFO, "param: k=%d, n=%d, bsize=%d\n",
              param_ksize, param_nsize, param_bsize);
          return 0;
          break;
        }
      }
      st = i+1;
    }

    if (param[i] == '\0')
      break;
  }

  return -1;
}


/*
 * extension initalization function
 *
 * return 0 on success, -1 or nonzero value on error
 */
int init()
{
  // initalization codes here

  // register function must always be the last call
  ipqext_register(&ext);

  return 0;
}

/*
 * extension finalization function
 *
 * return 0 on success, -1 or nonzero value on error
 */
int fini()
{
  // unregister function must always be the first call
  ipqext_unregister(&ext);

  // finialization codes here

  return 0;
}

/*
 * module initalization function
 */
void _init(void)
{
}

/*
 * module finalization function
 */
void _fini(void)
{
}

//----------------------------------------------------------------------//
