#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <time.h>
#include <signal.h>
#include <arpa/inet.h>
#include <lib/util.h>
#include <math.h>

struct sockaddr_in cliAddr;
int servSd = -1;
unsigned int totalcount, count;
char buffer[1400];
int iter;
int seqno;
int errcnt;

/* initialize server socket */
static int init()
{
  struct sockaddr_in servAddr;
  int one = 1;

  servSd = socket(PF_INET, SOCK_DGRAM, 0);
  if (servSd < 0)
    ERR_RET(-1);

  setsockopt(servSd, SOL_SOCKET, SO_REUSEADDR, (char *) &one, sizeof(one));
  servAddr.sin_family = AF_INET;
  servAddr.sin_port = htons(0);
  servAddr.sin_addr.s_addr = INADDR_ANY;
  if (bind(servSd, (struct sockaddr *) &servAddr, sizeof(servAddr)) < 0)
    ERR_RET(-1);

  return 0;
}

/* close server socket */
static inline void close_socket()
{
  if (servSd >= 0)
    close(servSd);
}


/* clean up process */
static void cleanup()
{
  close_socket();
}

/* timer event */
static void timer_event(int signum)
{
  int i, ret;
  int size = random() % sizeof(buffer);
  if (size < 100)
    size = 100;
//  printf("send packet with size = %d\n", size);
  for (i = 0; i < iter; i++) {
    ((unsigned int *) buffer)[0] = htonl(seqno);
    ret = sendto(servSd, buffer, size, 0,
                 (struct sockaddr *) &cliAddr, sizeof(cliAddr));
    if (ret == -1)
      errcnt++;

    seqno++;
    if (++count == totalcount) {
      printf("number of packet sent = %d\n", count);
      printf("number of errors = %d\n", errcnt);
      exit(0);
    }
  }
}

/* signal handler */
static void signal_handler(int signo)
{
  exit(0);
}

/* catch signal */
static inline void catch_signal()
{
  signal(SIGINT, signal_handler);       // catch CTRL-C signal
  signal(SIGQUIT, signal_handler);      // catch CTRL-\ signal
}

/* clear signal */
static inline void clear_signal()
{
  signal(SIGINT, SIG_DFL);      // restore CTRL-C signal
  signal(SIGQUIT, SIG_DFL);     // restore CTRL-\ signal
}

/* exit handler */
static void exit_handler()
{
  cleanup();
}

/* init timer */
static void init_timer(int rate, int srate)
{
  struct itimerval itv;
  double period;
  long sec, usec;

  iter = rate;

  period = 1.0 / (double) srate;
  sec = (long) period;
  usec = (long) (period * 1000000) % 1000000;

  printf("rate = %d packet/sec   period = %lds %ldus\n", rate * srate, sec,
         usec);
  itv.it_interval.tv_sec = sec;
  itv.it_interval.tv_usec = usec;
  itv.it_value = itv.it_interval;
  signal(SIGALRM, timer_event);
  setitimer(ITIMER_REAL, &itv, NULL);
}

/* main */
int main(int argc, char **argv)
{
  if (argc != 6) {
    printf("Usage: %s <host> <port> <num of packet> <rate> <srate>\n",
           argv[0]);
    exit(1);
  }

  atexit(&exit_handler);
  catch_signal();

  memset(buffer, 0x00, sizeof(buffer));

  cliAddr.sin_family = AF_INET;
  cliAddr.sin_addr.s_addr = inet_addr(argv[1]);
  cliAddr.sin_port = htons(atoi(argv[2]));
  totalcount = atoi(argv[3]);
  count = 0;
  seqno = 0;
  errcnt = 0;

  if (init() == -1)
    exit(1);

  srandom(time(NULL));
  init_timer(atoi(argv[4]), atoi(argv[5]));     // packets per sec

  for (;;);

  cleanup();

  return 0;
}
